@@ -3,7 +3,7 @@
 /*
     This file is part of Magnum.
 
-    Copyright © 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018
+    Copyright © 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019
               Vladimír Vondruš <mosra@centrum.cz>
 
     Permission is hereby granted, free of charge, to any person obtaining a
@@ -161,6 +161,35 @@ Prints the value with 4 significant digits.
 */
 MAGNUM_EXPORT Corrade::Utility::Debug& operator<<(Corrade::Utility::Debug& debug, Half value);
 
+namespace Implementation {
+
+template<> struct StrictWeakOrdering<Half> {
+    bool operator()(Half a, Half b) const {
+        /* Not mathematically equivalent to <, but does order */
+        return a.data() < b.data();
+    }
+};
+
+}
+
 }}
 
+#if defined(DOXYGEN_GENERATING_OUTPUT) || defined(CORRADE_TARGET_UNIX) || (defined(CORRADE_TARGET_WINDOWS) && !defined(CORRADE_TARGET_WINDOWS_RT)) || defined(CORRADE_TARGET_EMSCRIPTEN)
+namespace Corrade { namespace Utility {
+
+/**
+@tweakableliteral{Magnum::Math::Half}
+
+Parses the @link Magnum::Math::Literals::operator""_h @endlink literal.
+*/
+template<> struct MAGNUM_EXPORT TweakableParser<Magnum::Math::Half> {
+    TweakableParser() = delete;
+
+    /** @brief Parse the value */
+    static std::pair<TweakableState, Magnum::Math::Half> parse(Containers::ArrayView<const char> value);
+};
+
+}}
+#endif
+
 #endif
